/*
 * Copyright (C) 2003 Ichiro Satoh (ichiro@nii.ac.jp)
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting
 * documentation.
 *
 * FREE USE OF THIS SOFTWARE IS ALLOWED IN ITS "AS IS" CONDITION.
 * BUT WITHOUT WARRANTY. THE AUTHORS DISCLAIM ANY LIABILITY OF ANY 
 * KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS 
 * SOFTWARE.
 */

package tinyrmi;

import java.io.Serializable;
import java.io.IOException;
import java.util.Vector;

public class Message implements Serializable {
  protected String name  = null;
  protected int size = 0;
  protected ObjectId oid = null;
  private Object[] objects = null;
  private Class[] classes = null;
  public transient Vector args  = null;
  public transient Vector types = null;
  public transient Object source = null;

  public Message() {
    args = new Vector();
    types = new Vector();
  }

  public Message(String name) {
    this();
    this.name = name;
  }

  public Message(ObjectId oid, String name) {
    this(name);
    this.oid = oid;
  }

  public Message(ObjectId oid, String name, Object obj) {
    this(oid, name);
    setArg(obj);
  }

  public void setSource(Object source) {
    this.source = source;
  }

  public Object getSource() {
    return source;
  }

  public void setArg(Object obj) {
    types.addElement(new Byte(Type.TYPE_OBJECT));
    args.addElement(obj);
    size = size + 1;
  }

  public void setArg(boolean value) {
    types.addElement(new Byte(Type.TYPE_BOOLEAN));
    args.addElement(new Boolean(value));
    size = size + 1;
  }

  public void setArg(byte value) {
    types.addElement(new Byte(Type.TYPE_BYTE));
    args.addElement(new Byte(value));
    size = size + 1;
  }

  public void setArg(short value) {
    types.addElement(new Byte(Type.TYPE_SHORT));
    args.addElement(new Short(value));
    size = size + 1;
  }

  public void setArg(char value) {
    types.addElement(new Byte(Type.TYPE_CHAR));
    args.addElement(new Character(value));
    size = size + 1;
  }

  public void setArg(int value) {
    types.addElement(new Byte(Type.TYPE_INT));
    args.addElement(new Integer(value));
    size = size + 1;
  }

  public void setArg(long value) {
    types.addElement(new Byte(Type.TYPE_LONG));
    args.addElement(new Long(value));
    size = size + 1;
  }

  public void setArg(float value) {
    types.addElement(new Byte(Type.TYPE_FLOAT));
    args.addElement(new Float(value));
    size = size + 1;
  }

  public void setArg(double value) {
    types.addElement(new Byte(Type.TYPE_DOUBLE));
    args.addElement(new Double(value));
    size = size + 1;
  }

  public int getArgCount() {
    return size;
  }

  public void setName(String name) {
    this.name = name;
  }

  public String getName() {
    return name;
  }

  public void setID(ObjectId oid) {
    this.oid = oid;
  }

  public ObjectId getID() {
    return oid;
  }

  public String toString() {
    getArgs();
    String s = "[" + getID() + ":" + getName() + "|";
    for (int i = 0 ; i < getArgCount() ; i++ ) {
      Object obj = objects[i];
      s = s + ", " + obj.getClass();
    }
    s = s + "]";
    return s;
  }

  public Object getArg(int i) {
    return objects[i];
  }
   
  public Object[] getArgs() {
    if (objects == null) {
      objects = new Object[size];
      for (int i = 0; i < size; i++) {
	objects[i] = args.elementAt(i);
      }
    }
    return objects;
  }
   
  public Class[] getTypes() {
    if (classes == null) {
      classes = new Class[size];
      for (int i = 0; i < size; i++) {
				switch (((Byte)types.elementAt(i)).byteValue()) {
				case Type.TYPE_OBJECT:
					classes[i] = args.elementAt(i).getClass();
					break;
				case Type.TYPE_BOOLEAN:
					classes[i] = java.lang.Boolean.TYPE;
					break;
				case Type.TYPE_INT:
					classes[i] = java.lang.Integer.TYPE;
					break;
				case Type.TYPE_SHORT:
					classes[i] = java.lang.Short.TYPE;
					break;
				case Type.TYPE_LONG:
					classes[i] = java.lang.Long.TYPE;
					break;
				case Type.TYPE_FLOAT:
					classes[i] = java.lang.Float.TYPE;
					break;
				case Type.TYPE_DOUBLE:
					classes[i] = java.lang.Double.TYPE;
					break;
				case Type.TYPE_VOID:
					classes[i] = java.lang.Void.TYPE;
					break;
				default:
					throw new UnknownError("Invalid parameter type");	
				}
      }
    }
    return classes;
  }

  private synchronized void writeObject(java.io.ObjectOutputStream out)
  throws IOException {
    out.writeObject(oid);
    out.writeObject(name);
    out.writeInt(size);
    for (int i = 0; i < size; i++) {
      Object obj = args.elementAt(i);
      switch (((Byte)types.elementAt(i)).byteValue()) {
      case Type.TYPE_OBJECT:
	out.writeByte(Type.TYPE_OBJECT);
	out.writeObject(obj);
	break;
      case Type.TYPE_BOOLEAN:
	out.writeByte(Type.TYPE_BOOLEAN);
	out.writeObject(obj);
	break;
      case Type.TYPE_BYTE:
	out.writeByte(Type.TYPE_BYTE);
	out.writeObject(obj);
	break;
      case Type.TYPE_SHORT:
	out.writeByte(Type.TYPE_SHORT);
	out.writeObject(obj);
	break;
      case Type.TYPE_CHAR:
	out.writeByte(Type.TYPE_CHAR);
	out.writeObject(obj);
	break;
      case Type.TYPE_INT:
	out.writeByte(Type.TYPE_INT);
	out.writeObject(obj);
	break;
      case Type.TYPE_LONG:
	out.writeByte(Type.TYPE_LONG);
	out.writeObject(obj);
	break;
      case Type.TYPE_FLOAT:
	out.writeByte(Type.TYPE_FLOAT);
	out.writeObject(obj);
	break;
      case Type.TYPE_DOUBLE:
	out.writeByte(Type.TYPE_DOUBLE);
	out.writeObject(obj);
	break;
      case Type.TYPE_VOID:
	out.writeByte(Type.TYPE_VOID);
	out.writeObject(obj);
	break;
      default:
	throw new UnknownError("Invalid parameter type");	
      }
    }
  }

  private synchronized void readObject(java.io.ObjectInputStream in)
  throws IOException, ClassNotFoundException {
    oid = (ObjectId)in.readObject();
    name = (String)in.readObject();
    size = in.readInt();
    classes = new Class[size];
    objects = new Object[size];
    for (int i = 0 ; i < size ; i++ ) {
      byte type  = in.readByte();
      Object anObject = (Object)in.readObject();
      switch (type) {
      case Type.TYPE_OBJECT:
	classes[i] = anObject.getClass();
	objects[i] = anObject;
	break;
      case Type.TYPE_BOOLEAN:
	classes[i] = java.lang.Boolean.TYPE;
	objects[i] = (Boolean)anObject;
	break;
      case Type.TYPE_BYTE:
	classes[i] = java.lang.Byte.TYPE;
	objects[i] = (Byte)anObject;
	break;
      case Type.TYPE_SHORT:
	classes[i] = java.lang.Short.TYPE;
	objects[i] = (Short)anObject;
	break;
      case Type.TYPE_CHAR:
	classes[i] = java.lang.Character.TYPE;
	objects[i] = (Character)anObject;
	break;
      case Type.TYPE_INT:
	classes[i] = java.lang.Integer.TYPE;
	objects[i] = (Integer)anObject;
	break;
      case Type.TYPE_LONG:
	classes[i] = java.lang.Long.TYPE;
	objects[i] = (Long)anObject;
	break;
      case Type.TYPE_FLOAT:
	classes[i] = java.lang.Float.TYPE;
	objects[i] = (Float)anObject;
	break;
      case Type.TYPE_DOUBLE:
	classes[i] = java.lang.Double.TYPE;
	objects[i] = (Double)anObject;
	break;
      case Type.TYPE_VOID:
	classes[i] = java.lang.Void.TYPE;
	objects[i] = (Void)anObject;
	break;
      default:
	throw new UnknownError("Invalid parameter type");	
      }
    }
  }
}


