/*
 * Copyright (C) 2003 Ichiro Satoh (ichiro@nii.ac.jp)
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and
 * its documentation is hereby granted, provided that both the
 * copyright notice and this permission notice appear in all copies of
 * the software, derivative works or modified versions, and any
 * portions thereof, and that both notices appear in supporting
 * documentation.
 *
 * FREE USE OF THIS SOFTWARE IS ALLOWED IN ITS "AS IS" CONDITION.
 * BUT WITHOUT WARRANTY. THE AUTHORS DISCLAIM ANY LIABILITY OF ANY 
 * KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS 
 * SOFTWARE.
 */
 
package tinyrmi.network;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.Hashtable;

public class Channel implements Runnable {
	static final int REQUEST = 0x01;
	static final int RESPONSE = 0x02;
	static final int CLOSE = 0x43;
	static final int DISCONNECT = 0x05;

	protected ChannelManager cm = null;
	protected ObjectInputStream in = null;
	protected ObjectOutputStream out = null;
	protected OutputStream outStream;
	protected InputStream inStream;
	private Object lock = new Object();
	private int requestID = 0;
	private int responseID = 0;
	private Hashtable listeners;
	private Hashtable registry;

	public Channel(InputStream inStream, OutputStream outStream)
		throws IOException {
		this.inStream = inStream;
		this.outStream = outStream;
		registry = new Hashtable();
		listeners = new Hashtable();
		cm = new ChannelManager(this);
	}

	public void run() {
		try {
			if (in == null) {
				in = new ObjectInputStream(inStream);
			}
			if (out == null) {
				out = new ObjectOutputStream(outStream);
			}
			while (true) {
				int command = in.readByte();
				switch (command) {
				case REQUEST: {
					int id = in.readInt();
					Request req = (Request) in.readObject();
					//	  System.out.println("Channel: REQUEST("+id+") = " + req);
					RequestHandler handler = new RequestHandler(id, req);
					(new Thread(handler)).start();
					break;
				}
				case RESPONSE: {
					int id = in.readInt();
					Response res = (Response) in.readObject();
					//	  System.out.println("Channel: RESPONSE("+id+") = " + res);
					Request req = (Request) registry.get(new Integer(id));
					if (req != null) {
						synchronized (lock) {
							req.setResponse(res);
							req.setFinished(true);
							lock.notifyAll();
						}
					}
					break;
				}
				case CLOSE: {
					int id = in.readInt();
					synchronized (out) {
						out.writeInt(requestID);
						out.writeInt(DISCONNECT);
						out.flush();
					}
					break;
				}
				case DISCONNECT: {
					int id = in.readInt();
					synchronized (out) {
						in.close();
						out.close();
					}
					break;
				}
				default :
					break;
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
			try {
				outStream.close();
				inStream.close();
			}
			catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}

	public Response call(Request req) throws IOException {
		try {
			if (out == null) {
				out = new ObjectOutputStream(outStream);
			}
			int id;
			synchronized (out) {
				requestID = (++requestID) & 0x7FFF;
				id = requestID;
				registry.put(new Integer(id), req);
				out.writeByte(REQUEST);
				out.writeInt(id);
				out.writeObject(req);
				out.flush();
				//				System.out.println("Channel: call(" + req + ") id = " + id);
			}
			synchronized (lock) {
				while (!req.getFinished()) {
					try {
						lock.wait();
					}
					catch (InterruptedException e) {}
				}
				registry.remove(new Integer(id));
			}
		}
		catch (IOException e) {
			e.printStackTrace();
			throw e;
		}
		return req.getResponse();
	}

	public void close() throws IOException {
		try {
			requestID = (++requestID) & 0x7FFF;
			synchronized (out) {
				out.writeInt(CLOSE);
				out.writeInt(requestID);
				out.flush();
			}
		}
		catch (IOException e) {
			e.printStackTrace();
			throw e;
		}
	}

	class RequestHandler implements Runnable {
		Request req = null;
		int id;
		boolean sync = true;

		public RequestHandler(int id, Request req) {
			this.id = id;
			this.req = req;
		}

		public void run() {
			try {
				Response res = cm.handle(req);
				synchronized (out) {
					out.writeByte(RESPONSE);
					out.writeInt(id);
					out.writeObject(res);
					out.flush();
					//  System.out.println("ChannelRequestHandler: RESPONSE sent out = " + res);
				}
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
	}
}
