package tinyrmi.runtime;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InvalidClassException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OptionalDataException;
import java.io.StreamCorruptedException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Hashtable;

import tinyrmi.Message;
import tinyrmi.NotBoundException;
import tinyrmi.runtime.loader.InvocationException;

public class RuntimeImpl implements Runtime {
  static private final String DefaultLoaderHandler = 
    "runtime.loader.ObjectHandler";
  protected Hashtable listeners;
	ClassLoader loader = null;
	Object obj = null;
	Class cls = null;

	public RuntimeImpl() {
    listeners = new Hashtable();
	}

	public RuntimeImpl(ClassLoader loader) {
		this();
		this.loader = loader;
	}

	public void setClassLoader(ClassLoader loader) {
		this.loader = loader;
	}

	public ClassLoader getClassLoader() {
		if (loader == null) {
			loader = this.getClass().getClassLoader();
		}
		return loader;
	}

	public Object getObject() {
		return obj;
	}

	public Object newInstance(Message msg)
		throws InstantiationException, ClassNotFoundException {
		try {
			cls = getClassLoader().loadClass(msg.getName());
			Constructor cst = cls.getConstructor( msg.getTypes());
			obj = cst.newInstance(msg.getArgs());
		}
    catch (Exception e) {
      e.printStackTrace();
			throw new InstantiationException();
    }
		return obj;
	}
	
	public Object invoke(Message msg)
		throws NoSuchMethodException, NotBoundException {
		if (obj == null) {
			throw new NotBoundException("No Such Object");
		}
		try {
			Method meth = getMethod(cls, msg.getName(), msg.getTypes(), false);
			if (meth == null) {
				throw new NoSuchMethodException("No Such Method: "+msg.getName());
			}
			return meth.invoke(obj, msg.getArgs());
		}
    catch (InvocationTargetException e) {
      e.printStackTrace();
			throw new NoSuchMethodException(e.getMessage());
    }
    catch (IllegalAccessException e) {
      e.printStackTrace();
    }
		return null;
	}

  protected Method getMethod(Class targetClass, String methodName, 
																		Class[] classes, boolean asStatic) 
		throws NoSuchMethodException {
    int length = 0;
    if (classes != null) {
      length = classes.length;
    }
    Method[] meths = targetClass.getMethods();
    for( int i = 0 ; i < meths.length ; i++ ) {
      if (meths[i].getName().equals(methodName) &&
          meths[i].getParameterTypes().length == length &&
          (!asStatic || (meths[i].getModifiers() & Modifier.STATIC) != 0)) {
        Class[] methodArgClasses = meths[i].getParameterTypes();
        boolean isMatch = true;
        for(int j = 0 ; isMatch && j < length ; j++ ) {
          isMatch = isMatch && 
						(classes[j] == null ||
						 methodArgClasses[j].equals(classes[j]) ||
						 methodArgClasses[j].isAssignableFrom(classes[j]));
        }
        if (isMatch) {
          return meths[i];
				}
      }
    }
    throw new NoSuchMethodException(methodName + " in " + targetClass);
  }

	public Object resume(Message msg)
		throws ClassNotFoundException, StreamCorruptedException, 
					 OptionalDataException {
		if (msg.getArgs().length > 0) {
			try {
				byte[] data = (byte[])msg.getArg(0);
				ByteArrayInputStream bytein = new ByteArrayInputStream(data);
				ObjectInputStream objin = new ObjectInputStream(bytein);
				obj = objin.readObject();
				cls = obj.getClass();
				objin.close();
				bytein.close();
			}
			catch (ClassNotFoundException e) {
				e.printStackTrace();
			}
			catch (InvalidClassException e) {
				e.printStackTrace();
			}
			catch (StreamCorruptedException e) {
				e.printStackTrace();
			}
			catch (OptionalDataException e) {
				e.printStackTrace();
			}
			catch (IOException e) {
				e.printStackTrace();
			}
		}
		return obj;
	}
	
  public byte[] suspend(Message msg) {
    byte[] data = null;
    try {
      ByteArrayOutputStream byteout = new ByteArrayOutputStream();
      ObjectOutputStream objout = new ObjectOutputStream(byteout);
      objout.writeObject(obj);
      objout.flush();
      data = byteout.toByteArray();
      objout.close();
      byteout.close();
    } 
    catch (IOException e) {
      e.printStackTrace();
    }
    return data;
  }
	
  public void destroy() {}

  public Hashtable getListeners() {
    return listeners;
  }

  public Object addListener(Message msg) 
    throws NotBoundException, InvocationException {
		return null;
	}
}
